@@ -1,63 +1,61 @@
 # Copyright 2022 Touca, Inc. Subject to Apache-2.0 License.
 
+from dataclasses import dataclass
 from datetime import datetime, timedelta
 from enum import Enum
-from typing import Dict, Tuple
 from pathlib import Path
+from typing import Dict, Tuple
 
-from touca._types import IntegerType, ToucaType, VectorType, BlobType
+from touca._rules import ComparisonRule
+from touca._types import Artifact, BlobType, IntegerType, ToucaType, VectorType
 
 
 class ResultCategory(Enum):
-    """ """
-
     Check = 1
     Assert = 2
 
 
+@dataclass
 class ResultEntry:
     """
     Wrapper around a given ``ToucaType`` value that includes the category
     it should belong to.
-
-    We are intentionally not using ``@dataclass`` to ensure the core library
-    has no dependency on ``dataclasses`` module. This may change in the future.
     """
 
-    def __init__(self, typ: ResultCategory, val: ToucaType):
-        """
-        Creates an entry given its value and the category it should belong to.
-
-        :param typ: type of the entry
-        :param val: value of the entry
-        """
-        self.typ = typ
-        self.val = val
+    typ: ResultCategory
+    val: ToucaType
+    rule: ComparisonRule = None
 
 
 class Case:
-    """ """
-
     def __init__(self, **kwargs):
-
         self._meta = kwargs
-        self._artifacts: Dict[str, Path] = dict()
         self._results: Dict[str, ResultEntry] = dict()
         self._tics: Dict[str, datetime] = dict()
         self._tocs: Dict[str, datetime] = dict()
 
-    def check(self, key: str, value: ToucaType):
+    def check(self, key: str, value: ToucaType, rule: ComparisonRule = None):
         """
-        Logs a given value as a test result for the declared test case
+        Captures the value of a given variable as a data point for the declared
+        test case and associates it with the specified key.
+
+        :param key: name to be associated with the captured data point
+        :param value: value to be captured as a test result
+        :param rule: comparison rule for this test result
+        """
+        self._results[key] = ResultEntry(ResultCategory.Check, value, rule)
+
+    def check_file(self, key: str, file: Path):
+        """
+        Captures an external file as a data point for the declared test case
         and associates it with the specified key.
 
-        :param key: name to be associated with the logged test result
-        :param value: value to be logged as a test result
+        :param key: name to be associated with captured file
+        :param path: path to the external file to be captured
         """
-        self._results[key] = ResultEntry(typ=ResultCategory.Check, val=value)
-
-    def check_file(self, key: str, file):
-        self._artifacts[key] = file
+        self._results[key] = ResultEntry(
+            typ=ResultCategory.Check, val=BlobType(Artifact.from_file(file))
+        )
 
     def assume(self, key: str, value: ToucaType):
         """
@@ -223,15 +221,15 @@ class Case:
         }
 
     def json(self):
-        results = [
-            {"key": k, "value": v.val.json()}
-            for k, v in self._results.items()
-            if v.typ is ResultCategory.Check
-        ]
-        results.extend({"key": k, "value": str(v)} for k, v in self._artifacts.items())
         return {
             "metadata": self._metadata(),
-            "results": results,
+            "results": [
+                {"key": k, "value": v.val.json(), "rule": v.rule.json()}
+                if v.rule
+                else {"key": k, "value": v.val.json()}
+                for k, v in self._results.items()
+                if v.typ is ResultCategory.Check
+            ],
             "assertions": [
                 {"key": k, "value": v.val.json()}
                 for k, v in self._results.items()
@@ -241,14 +239,14 @@ class Case:
         }
 
     def serialize(self) -> bytearray:
-        import touca._schema as schema
+        import touca_fbs as schema
         from flatbuffers import Builder
 
         dicts = {
             ResultCategory.Check: schema.ResultType.Check,
             ResultCategory.Assert: schema.ResultType.Assert,
         }
-        builder = Builder(1024)
+        builder = Builder()
 
         metadata = {k: builder.CreateString(v) for k, v in self._metadata().items()}
         schema.MetadataStart(builder)
@@ -262,20 +260,12 @@ class Case:
         result_entries = []
         for k, v in self._results.items():
             fbs_key = Builder.CreateString(builder, k)
-            fbs_value = v.val.serialize(builder)
+            fbs_value = v.val.add_rule(v.rule).serialize(builder)
             schema.ResultStart(builder)
             schema.ResultAddKey(builder, fbs_key)
             schema.ResultAddValue(builder, fbs_value)
             schema.ResultAddTyp(builder, dicts.get(v.typ))
             result_entries.append(schema.ResultEnd(builder))
-        for k, v in self._artifacts.items():
-            fbs_key = Builder.CreateString(builder, k)
-            fbs_value = BlobType(v).serialize(builder)
-            schema.ResultStart(builder)
-            schema.ResultAddKey(builder, fbs_key)
-            schema.ResultAddValue(builder, fbs_value)
-            schema.ResultAddTyp(builder, schema.ResultType.Check)
-            result_entries.append(schema.ResultEnd(builder))
 
         schema.ResultsStartEntriesVector(builder, len(result_entries))
         for item in reversed(result_entries):
